import json

from jet_bridge import encoders


class Response(object):
    headers = {'Content-Type': 'application/json'}
    encoder_class = encoders.JSONEncoder

    def __init__(self, data=None, status=None, headers=None, exception=False, content_type=None):
        self.data = data
        self.status = status
        self.exception = exception
        self.content_type = content_type

        if headers:
            self.headers.update(headers)

    def header_items(self):
        if not self.headers:
            return []
        return self.headers.items()

    def render(self):
        if self.data is None:
            return bytes()

        return json.dumps(
            self.data,
            cls=self.encoder_class
        )



import os
from PIL import Image

from jet_bridge import settings
from jet_bridge.exceptions.not_found import NotFound
from jet_bridge.media_cache import cache
from jet_bridge.views.base.api import APIView


class ImageResizeHandler(APIView):

    def create_thumbnail(self, path, thumbnail_path, max_width, max_height):
        img = Image.open(path)
        img.thumbnail((max_width, max_height), Image.ANTIALIAS)

        if not os.path.exists(os.path.dirname(thumbnail_path)):
            try:
                os.makedirs(os.path.dirname(thumbnail_path))
            except OSError:
                raise

        img.save(thumbnail_path, format=img.format, quality=85)  # TODO: determine real extension from format

    def get(self):
        # TODO: Move to serializer

        path = self.get_argument('path')
        max_width = self.get_argument('max_width', 320)
        max_height = self.get_argument('max_height', 240)
        image_full_path = os.path.join(settings.MEDIA_ROOT, path)
        thumbnail_full_path = cache.full_path(path)

        if not os.path.exists(image_full_path):
            raise NotFound

        try:
            if not os.path.exists(thumbnail_full_path):
                cache.clear_cache_if_needed()
                self.create_thumbnail(image_full_path, thumbnail_full_path, max_width, max_height)
                cache.add_file(thumbnail_full_path)

            self.set_header('Content-Type', 'image/{}'.format(os.path.splitext(thumbnail_full_path)[1][1:]))

            with open(thumbnail_full_path, 'rb') as f:
                data = f.read()
                self.write(data)
            self.finish()
        except IOError as e:
            raise e



import os

from jet_bridge import settings
from jet_bridge.permissions import HasProjectPermissions
from jet_bridge.responses.base import Response
from jet_bridge.utils.common import get_random_string
from jet_bridge.views.base.api import APIView


class MediaFileUploadHandler(APIView):
    permission_classes = (HasProjectPermissions,)

    def get_available_name(self, name):
        dir_name, file_name = os.path.split(name)
        file_root, file_ext = os.path.splitext(file_name)

        while os.path.exists(name):
            name = os.path.join(dir_name, '%s_%s%s' % (file_root, get_random_string(7), file_ext))

        return name

    def post(self):
        # TODO: Move to serializer
        file = self.request.files.get('file', [])[0]
        path = self.get_body_argument('path')
        filename = self.get_body_argument('filename', file.filename)

        upload_path = os.path.join(settings.MEDIA_ROOT, path, filename)
        upload_path = self.get_available_name(upload_path)

        relative_upload_path = upload_path[len(settings.MEDIA_ROOT) + 1:]

        # TODO: Add validations

        if not os.path.exists(os.path.dirname(upload_path)):
            try:
                os.makedirs(os.path.dirname(upload_path))
            except OSError:
                raise

        with open(upload_path, 'wb') as f:
            f.write(file['body'])

        response = Response({
            'uploaded_path': relative_upload_path,
            'uploaded_url': self.build_absolute_uri('/media/{}'.format(relative_upload_path))
        })
        self.write_response(response)



from jet_bridge.serializers.serializer import Serializer


class ModelSerializer(Serializer):

    def __init__(self, *args, **kwargs):
        self.session = kwargs.get('context', {}).get('session', None)
        super(ModelSerializer, self).__init__(*args, **kwargs)

    def create(self, validated_data):
        ModelClass = self.meta.model

        instance = ModelClass(**validated_data)
        self.session.add(instance)
        self.session.commit()

        return instance

    def update(self, instance, validated_data):
        for attr, value in validated_data.items():
            setattr(instance, attr, value)
        self.session.commit()

        return instance



from sqlalchemy import sql, desc

from jet_bridge.filters.char_filter import CharFilter
from jet_bridge.filters.filter import EMPTY_VALUES


class OrderFilter(CharFilter):

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        if len(value) < 2:
            return qs.filter(sql.false())

        descending = value[0:1] == '-'
        value = value[1:] if descending else value
        entity = qs._primary_entity.entity_zero_or_selectable.entity
        column = getattr(entity, value, None)

        if column is None:
            return qs.filter(sql.false())

        if descending:
            column = desc(column)

        return qs.order_by(column)



import json

from jet_bridge.fields.field import Field


class JSONField(Field):

    def __init__(self, *args, **kwargs):
        super(JSONField, self).__init__(*args, **kwargs)

    def to_internal_value_item(self, value):
        return json.loads(value)

    def to_representation_item(self, value):
        return value



from jet_bridge.responses.base import Response


class ListAPIViewMixin(object):

    def list(self, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(instance=page, many=True)
            response = self.get_paginated_response(serializer.representation_data)
            self.write_response(response)
            return

        serializer = self.get_serializer(instance=queryset, many=True)
        response = Response(serializer.representation_data)
        self.write_response(response)



from sqlalchemy import inspect

from jet_bridge.filters.model import get_model_filter_class
from jet_bridge.filters.model_aggregate import ModelAggregateFilter
from jet_bridge.filters.model_group import ModelGroupFilter
from jet_bridge.permissions import HasProjectPermissions, ModifyNotInDemo
from jet_bridge.responses.base import Response
from jet_bridge.router import action
from jet_bridge.serializers.model import get_model_serializer
from jet_bridge.serializers.model_group import ModelGroupSerializer
from jet_bridge.views.mixins.model import ModelAPIViewMixin
from jet_bridge.db import MappedBase


class ModelHandler(ModelAPIViewMixin):
    model = None
    permission_classes = (HasProjectPermissions, ModifyNotInDemo)

    def prepare(self):
        mapper = inspect(self.get_model())
        self.lookup_field = mapper.primary_key[0].name

    @property
    def required_project_permission(self):
        return {
            'permission_type': 'model',
            'permission_object': self.path_kwargs['model'],
            'permission_actions': {
                'create': 'w',
                'update': 'w',
                'partial_update': 'w',
                'destroy': 'd',
                'retrieve': 'r',
                'list': 'r',
                'aggregate': 'r',
                'group': 'r',
                'reorder': 'w',
                'reset_order': 'w'
            }.get(self.action, 'w')
        }

    def get_model(self):
        if self.model:
            return self.model

        self.model = MappedBase.classes[self.path_kwargs['model']]

        return self.model

    def get_serializer_class(self):
        Model = self.get_model()
        return get_model_serializer(Model)

    def get_filter_class(self):
        return get_model_filter_class(self.get_model())

    def get_queryset(self):
        Model = self.get_model()

        return self.session.query(Model)

    @action(methods=['get'], detail=False)
    def aggregate(self, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        y_func = self.get_argument('_y_func').lower()
        y_column = self.get_argument('_y_column', self.lookup_field)

        model_serializer = self.get_serializer()

        y_serializers = list(filter(lambda x: x.field_name == y_column, model_serializer.fields))
        y_serializer = y_serializers[0]

        queryset = ModelAggregateFilter().filter(queryset, {
            'y_func': y_func,
            'y_column': y_column
        })

        result = y_serializer.to_representation(queryset[0])  # TODO: Refactor serializer

        self.write_response(Response({
            'y_func': result
        }))

    @action(methods=['get'], detail=False)
    def group(self, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        x_column = self.get_argument('_x_column')
        x_lookup_name = self.get_argument('_x_lookup')
        y_func = self.get_argument('_y_func').lower()
        y_column = self.get_argument('_y_column', self.lookup_field)

        model_serializer = self.get_serializer()

        x_serializers = list(filter(lambda x: x.field_name == x_column, model_serializer.fields))
        x_serializer = x_serializers[0]

        y_serializers = list(filter(lambda x: x.field_name == y_column, model_serializer.fields))
        y_serializer = y_serializers[0]

        queryset = ModelGroupFilter().filter(queryset, {
            'x_column': x_column,
            'x_lookup': x_lookup_name,
            'y_func': y_func,
            'y_column': y_column
        })
        serializer = ModelGroupSerializer(
            instance=queryset,
            many=True,
            # TODO: Refactor serializer
            # group_serializer=x_serializer,
            # y_func_serializer=y_serializer
        )

        self.write_response(Response(serializer.representation_data))



from six.moves.urllib_parse import quote

from jet_bridge import settings
from jet_bridge.utils.backend import register_token
from jet_bridge.views.base.api import APIView


class RegisterHandler(APIView):

    def get(self, *args, **kwargs):
        token, created = register_token()

        if not token:
            return

        if settings.WEB_BASE_URL.startswith('https') and not self.request.full_url().startswith('https'):
            web_base_url = 'http{}'.format(settings.WEB_BASE_URL[5:])
        else:
            web_base_url = settings.WEB_BASE_URL

        url = '{}/projects/register/{}'.format(web_base_url, token.token)
        query_string = 'referrer={}'.format(quote(self.request.full_url().encode('utf8')))

        self.redirect('%s?%s' % (url, query_string))



from sqlalchemy import create_engine, MetaData
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import sessionmaker

from jet_bridge import settings
from jet_bridge.models import Base


def build_engine_url():
    url = [
        settings.DATABASE_ENGINE,
        '://'
    ]

    if settings.DATABASE_USER:
        url.append(settings.DATABASE_USER)

        if settings.DATABASE_PASSWORD:
            url.append(':')
            url.append(settings.DATABASE_PASSWORD)

        if settings.DATABASE_HOST:
            url.append('@')

    if settings.DATABASE_HOST:
        url.append(settings.DATABASE_HOST)

        if settings.DATABASE_PORT:
            url.append(':')
            url.append(str(settings.DATABASE_PORT))

        url.append('/')

    url.append(settings.DATABASE_NAME)

    return ''.join(url)

engine_url = build_engine_url()
engine = create_engine(engine_url)
Session = sessionmaker(bind=engine)

print('Connected to database engine "{}" with name "{}"'.format(settings.DATABASE_ENGINE, settings.DATABASE_NAME))

Base.metadata.create_all(engine)

metadata = MetaData()
metadata.reflect(engine)
MappedBase = automap_base(metadata=metadata)
MappedBase.prepare()



import sqlalchemy

from jet_bridge.filters.char_filter import CharFilter


FILTER_FOR_DBFIELD = {
    # sqlalchemy.AutoField:                   {'filter_class': NumberFilter},
    sqlalchemy.VARCHAR: {'filter_class': CharFilter},
    # sqlalchemy.TextField:                   {'filter_class': CharFilter},
    # sqlalchemy.BOOLEAN:                {'filter_class': BooleanFilter},
    # sqlalchemy.DateField:                   {'filter_class': DateFilter},
    # sqlalchemy.DateTimeField:               {'filter_class': DateTimeFilter},
    # sqlalchemy.TimeField:                   {'filter_class': TimeFilter},
    # sqlalchemy.DurationField:               {'filter_class': DurationFilter},
    # sqlalchemy.DecimalField:                {'filter_class': NumberFilter},
    # sqlalchemy.SmallIntegerField:           {'filter_class': NumberFilter},
    # sqlalchemy.IntegerField:                {'filter_class': NumberFilter},
    # sqlalchemy.PositiveIntegerField:        {'filter_class': NumberFilter},
    # sqlalchemy.PositiveSmallIntegerField:   {'filter_class': NumberFilter},
    # sqlalchemy.FloatField:                  {'filter_class': NumberFilter},
    # sqlalchemy.NullBooleanField:            {'filter_class': BooleanFilter},
    # sqlalchemy.SlugField:                   {'filter_class': CharFilter},
    # sqlalchemy.EmailField:                  {'filter_class': CharFilter},
    # sqlalchemy.FilePathField:               {'filter_class': CharFilter},
    # sqlalchemy.URLField:                    {'filter_class': CharFilter},
    # sqlalchemy.GenericIPAddressField:       {'filter_class': CharFilter},
    # sqlalchemy.CommaSeparatedIntegerField:  {'filter_class': CharFilter},
    # sqlalchemy.UUIDField:                   {'filter_class': UUIDFilter}
}
FILTER_FOR_DBFIELD_DEFAULT = {'filter_class': CharFilter}



import hashlib
import random
import time

try:
    random = random.SystemRandom()
    using_sysrandom = True
except NotImplementedError:
    import warnings
    warnings.warn('A secure pseudo-random number generator is not available '
                  'on your system. Falling back to Mersenne Twister.')
    using_sysrandom = False


def get_random_string(length, allowed_chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789', salt=''):
    """
    Returns a securely generated random string.

    The default length of 12 with the a-z, A-Z, 0-9 character set returns
    a 71-bit value. log_2((26+26+10)^12) =~ 71 bits
    """
    if not using_sysrandom:
        # This is ugly, and a hack, but it makes things better than
        # the alternative of predictability. This re-seeds the PRNG
        # using a value that is hard for an attacker to predict, every
        # time a random string is required. This may change the
        # properties of the chosen random sequence slightly, but this
        # is better than absolute predictability.
        random.seed(
            hashlib.sha256(
                ("%s%s%s" % (
                    random.getstate(),
                    time.time(),
                    salt)).encode('utf-8')
            ).digest())
    return ''.join(random.choice(allowed_chars) for i in range(length))



from sqlalchemy import func, sql

from jet_bridge.filters.char_filter import CharFilter
from jet_bridge.filters.filter import EMPTY_VALUES


class ModelAggregateFilter(CharFilter):

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        entity = qs._primary_entity.entity_zero_or_selectable.entity
        y_column = getattr(entity, value['y_column'])

        if value['y_func'] == 'count':
            y_func = func.count(y_column)
        elif value['y_func'] == 'sum':
            y_func = func.sum(y_column)
        elif value['y_func'] == 'min':
            y_func = func.min(y_column)
        elif value['y_func'] == 'max':
            y_func = func.max(y_column)
        elif value['y_func'] == 'avg':
            y_func = func.avg(y_column)
        else:
            return qs.filter(sql.false())

        qs = qs.session.query(y_func).one()

        return qs
